/*
 * Copyright (C) 2023 IOPSYS Software Solutions AB. All rights reserved.
 */
#if (EASYMESH_VERSION > 2)
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/stat.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>

#include <linux/if_bridge.h>
#include <netlink/socket.h>

#include <netinet/in.h>
#include <net/if.h>

#include <cmdu.h>
#include <1905_tlvs.h>
#include <easy/easy.h>
#include <easy/utils.h>
//#include "wifi.h"

#include <i1905_wsc.h>

#include <uci.h>
#include <map_module.h>
#include <easymesh.h>
#include <dirent.h>

#include <wifidefs.h>
#include "wifi_dataelements.h"

#include "timer.h"
#include "utils/utils.h"
#include "utils/debug.h"
#include "utils/dpp_sock.h"
#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
#include <dpp_api.h>
#endif

#include "dpp.h"
#include <math.h>
#endif
//#include "config.h"
#include "cntlr.h"
#include "cntlr_ubus.h"
#include "cntlr_ubus_dbg.h"
#include "cntlr_map.h"
#include "cntlr_cmdu.h"
#include "cntlr_tlv.h"

static void dpp_fd_timeout(atimer_t *t);

#ifndef USE_LIBDPP
uint8_t *dpp_get_attr(uint8_t *msg, uint16_t msglen, uint16_t attr, uint16_t *olen)
{
	uint8_t *p;
	uint8_t *msg_end;


	if (!msg || msglen == 0)
		return NULL;

	p = msg;
	msg_end = msg + msglen;

	while (labs(p - msg) < msglen - 4) {
		uint16_t attr_type;
		uint16_t attr_len;

		attr_type = buf_get_le16(p);
		p += 2;
		attr_len = buf_get_le16(p);
		p += 2;

		if (p + attr_len > msg_end) {
			pr_debug("%s: Invalid attr = 0x%04x  len = %hu, p = %p  msg_end = %p diff = %d msg_len:%d\n",
				__func__, attr_type, attr_len, p, msg_end, abs(msg_end - p), msglen);
			return NULL;
		}

		if (attr_type == attr) {
			pr_debug("MATCHED attr = 0x%04x\n", attr);
			if (olen)
				*olen = attr_len;

			return p;
		}

		p += attr_len;
	}

	pr_debug("%s: VALID msg, but attr 0x%04x not found!\n", __func__, attr);
	return NULL;
}

uint8_t dpp_get_pub_af_type(const uint8_t *frame, uint16_t framelen)
{
	if (framelen < 7)
		return 255;

	return *(frame + 6);
}

uint8_t dpp_get_gas_frame_type(const uint8_t *frame, uint16_t framelen)
{
	if (framelen < 1)
		return 255;

	return *frame;
}

uint8_t dpp_get_frame_type(const uint8_t *frame, uint16_t framelen)
{
	if (framelen < 7)
		return 255;

	if (FRAME_IS_DPP_PUB_AF(frame))
		return dpp_get_pub_af_type(frame, framelen);
	else if (FRAME_IS_DPP_GAS_FRAME(frame))
		return dpp_get_gas_frame_type(frame, framelen);

	return 255;
}

#endif

const char *dpp_type2str(enum dpp_bootstrap_type type)
{
	if (type == DPP_BOOTSTRAP_QR_CODE_WIFI)
		return "qrcode";
	else if (type == DPP_BOOTSTRAP_PKEX)
		return "pkex";
	else if (type == DPP_BOOTSTRAP_NFC_URI)
		return "nfc";
	return "unknown";
}

const char *dpp_state2str(enum dpp_responder_state state)
{
	if (state == DPP_BOOSTRAPPING)
		return "bootstrapping";
	else if (state == DPP_AUTHENTICATING)
		return "authenticating";
	else if (state == DPP_AUTHENTICATED)
		return "authenticated";
	return "unknown";
}

const char *dpp_frame_type2str(uint8_t type)
{
	if (type == DPP_PA_AUTHENTICATION_REQ)
		return "Authentication Request";
	else if (type == DPP_PA_AUTHENTICATION_RESP)
		return "Authentication Response";
	else if (type == DPP_PA_AUTHENTICATION_CONF)
		return "Authentication Confirm";
	else if (type == DPP_PA_PEER_DISCOVERY_REQ)
		return "Peer Discovery Request";
	else if (type == DPP_PA_PEER_DISCOVERY_RESP)
		return "Peer Discovery Response";
	else if (type == DPP_PA_PKEX_V1_EXCHANGE_REQ)
		return "PKEX Version 1 Exchange Request";
	else if (type == DPP_PA_PKEX_EXCHANGE_RESP)
		return "PKEX Exchange Response";
	else if (type == DPP_PA_PKEX_COMMIT_REVEAL_REQ)
		return "PKEX Commit-Reveal Request";
	else if (type == DPP_PA_PKEX_COMMIT_REVEAL_RESP)
		return "PKEX Commit-Reveal Response";
	else if (type == DPP_PA_CONFIGURATION_RESULT)
		return "Configuration Result";
	else if (type == DPP_PA_CONNECTION_STATUS_RESULT)
		return "Connection Status Result";
	else if (type == DPP_PA_PRESENCE_ANNOUNCEMENT)
		return "Presence Announcement";
	else if (type == DPP_PA_RECONFIG_ANNOUNCEMENT)
		return "Reconfiguration Announcement";
	else if (type == DPP_PA_RECONFIG_AUTH_REQ)
		return "Reconfiguration Authentication Request";
	else if (type == DPP_PA_RECONFIG_AUTH_RESP)
		return "Reconfiguration Authentication Response";
	else if (type == DPP_PA_RECONFIG_AUTH_CONF)
		return "Reconfiguration Authentication Confirm";
	else if (type == DPP_PA_PKEX_EXCHANGE_REQ)
		return "PKEX Exchange Request";
	else if (type == 0x13)
		return "Push Button Presence Announcement";
	else if (type == 0x14)
		return "Push Button Presence Announcement Response";
	else if (type == 0x15)
		return "Private Peer Introduction Query";
	else if (type == 0x16)
		return "Private Peer Introduction Notify";
	else if (type == 0x17)
		return "Private Peer Introduction Update";
	return "Unknown";
}

struct dpp_enrollee *dpp_enrollee_get(struct controller *c, uint8_t *mac)
{
	struct dpp_enrollee *e;

	list_for_each_entry(e, &c->dpp_ctx.enrolleelist, list) {
		if (!memcmp(e->mac, mac, 6))
			return e;
	}

	return NULL;
}

struct dpp_enrollee *dpp_enrollee_get_hash(struct controller *c, uint8_t *hash)
{
	struct dpp_enrollee *e;

	list_for_each_entry(e, &c->dpp_ctx.enrolleelist, list) {
		struct dpp_chirp *p = e->chirp;

		if (!memcmp(e->hash, hash, 32))
			return e;

		if (!p)
			continue;

		if (!memcmp(p->hash, hash, 32))
			return e;
	}

	return NULL;
}

/* PROXY AGENT MODE FUNC */
struct dpp_enrollee *dpp_enrollee_alloc(struct controller *c, uint8_t *mac)
{
	struct dpp_enrollee *e;

	e = calloc(1, sizeof(*e));
	if (!e) {
		err("|%s:%d| Out of memory!\n", __func__, __LINE__);
		return NULL;
	}

	e->band = BAND_UNKNOWN;
	e->c = c; /* TODO: deprecate */
	memcpy(e->mac, mac, 6);
	INIT_LIST_HEAD(&e->framelist);
	list_add(&e->list, &c->dpp_ctx.enrolleelist);

	return e;
}

void dpp_enrollee_free(struct dpp_enrollee *e)
{
	struct dpp_frame *f;

	list_for_each_entry(f, &e->framelist, list) {
		list_del(&f->list);
		free(f->data);
		free(f);
	}

	list_del(&e->list);
	free(e);
}

struct dpp_chirp *dpp_chirp_get(struct controller *c, uint8_t *hash)
{
	struct dpp_chirp *p;

	list_for_each_entry(p, &c->dpp_ctx.chirplist, list) {
		if (!memcmp(p->hash, hash, 32))
			return p;
	}

	return NULL;
}

/* PROXY AGENT MODE FUNC */
struct dpp_chirp *dpp_chirp_alloc(struct controller *c, uint8_t *hash)
{
	struct dpp_chirp *p;

	p = calloc(1, sizeof(*p));
	if (!p) {
		err("|%s:%d| Out of memory!\n", __func__, __LINE__);
		return NULL;
	}

	p->c = c;
	memcpy(p->hash, hash, 32);
	list_add(&p->list, &c->dpp_ctx.chirplist);
	timer_init(&p->timeout, dpp_fd_timeout);
	return p;
}

int dpp_enrollee_update_hash(struct controller *c,
			     struct dpp_enrollee *e,
			     uint8_t len, uint8_t *hash)
{
	struct dpp_chirp *p;

	if (e->chirp) {
		p = e->chirp;
		memcpy(p->hash, hash, 32);
		return 0;
	}

	p = dpp_chirp_get(c, hash);
	if (!p) {
		dbg("|%s:%d| no dpp chirp for hash, allocate\n",
			__func__, __LINE__);

		p = dpp_chirp_alloc(c, hash);
		if (!p) {
			dbg("|%s:%d| Failed to allocate enrollee for hash\n",
				__func__, __LINE__);
			return -1;
		}
	}

	e->chirp = p;
	return 0;
}

struct dpp_frame *dpp_enrollee_get_frame(struct controller *c,
					 struct dpp_enrollee *e, uint8_t type)
{
	struct dpp_frame *frame;

	list_for_each_entry(frame, &e->framelist, list) {
		if (type == frame->type)
			return frame;
	}

	return NULL;
}

struct dpp_frame *dpp_enrollee_add_frame(struct controller *c, struct dpp_enrollee *e,
			   uint8_t type, uint16_t len, uint8_t *data, uint8_t *almac)
{
	struct dpp_frame *frame;

	frame = dpp_enrollee_get_frame(c, e, type);
	if (!frame) {
		frame = calloc(1, sizeof(*frame));
		if (!frame) {
			err("|%s:%d| Out of memory!\n", __func__, __LINE__);
			return NULL;
		}

		frame->type = type;
		list_add(&frame->list, &e->framelist);
	} else {
		free(frame->data);
		frame->data = NULL;
		frame->len = 0;
	}

	frame->data = calloc(1, len);
	if (!frame->data) {
		err("|%s:%d| Out of memory!\n", __func__, __LINE__);
		list_del(&frame->list);
		free(frame);
		return NULL;
	}

	memcpy(frame->data, data, len);

	if (almac)
		memcpy(frame->proxy_almac, almac, 6);

	frame->len = len;
	return frame;
}
/* TODO: obvious gap 6GHz not handled! */
enum wifi_band wifi_channel_to_band(uint8_t channel)
{
	if (channel >= 1 && channel < 14)
		return BAND_2;
	else if (channel >= 36 && channel < 200)
		return BAND_5;

	return BAND_UNKNOWN;
}

static int ieee80211_chan_to_freq(uint8_t op_class, uint8_t chan)
{
	/* Table E-4 in IEEE Std 802.11-2012 - Global operating classes */
	switch (op_class) {
	case 81:
		/* channels 1..13 */
		if (chan < 1 || chan > 13)
			return -1;
		return 2407 + 5 * chan;
	case 82:
		/* channel 14 */
		if (chan != 14)
			return -1;
		return 2414 + 5 * chan;
	case 83: /* channels 1..9; 40 MHz */
	case 84: /* channels 5..13; 40 MHz */
		if (chan < 1 || chan > 13)
			return -1;
		return 2407 + 5 * chan;
	case 115: /* channels 36,40,44,48; indoor only */
	case 116: /* channels 36,44; 40 MHz; indoor only */
	case 117: /* channels 40,48; 40 MHz; indoor only */
	case 118: /* channels 52,56,60,64; dfs */
	case 119: /* channels 52,60; 40 MHz; dfs */
	case 120: /* channels 56,64; 40 MHz; dfs */
		if (chan < 36 || chan > 64)
			return -1;
		return 5000 + 5 * chan;
	case 121: /* channels 100-140 */
	case 122: /* channels 100-142; 40 MHz */
	case 123: /* channels 104-136; 40 MHz */
		if (chan < 100 || chan > 140)
			return -1;
		return 5000 + 5 * chan;
	case 124: /* channels 149,153,157,161 */
		if (chan < 149 || chan > 161)
			return -1;
		return 5000 + 5 * chan;
	case 125: /* channels 149,153,157,161,165,169,173,177 */
	case 126: /* channels 149,157,165,173; 40 MHz */
	case 127: /* channels 153,161,169,177; 40 MHz */
		if (chan < 149 || chan > 177)
			return -1;
		return 5000 + 5 * chan;
	case 128: /* center freqs 42, 58, 106, 122, 138, 155, 171; 80 MHz */
	case 130: /* center freqs 42, 58, 106, 122, 138, 155, 171; 80 MHz */
		if (chan < 36 || chan > 177)
			return -1;
		return 5000 + 5 * chan;
	case 129: /* center freqs 50, 114, 163; 160 MHz */
		if (chan < 36 || chan > 177)
			return -1;
		return 5000 + 5 * chan;
	case 131: /* UHB channels, 20 MHz: 1, 5, 9.. */
	case 132: /* UHB channels, 40 MHz: 3, 11, 19.. */
	case 133: /* UHB channels, 80 MHz: 7, 23, 39.. */
	case 134: /* UHB channels, 160 MHz: 15, 47, 79.. */
	case 135: /* UHB channels, 80+80 MHz: 7, 23, 39.. */
		if (chan < 1 || chan > 233)
			return -1;
		return 5950 + chan * 5;
	case 136: /* UHB channels, 20 MHz: 2 */
		if (chan == 2)
			return 5935;
		return -1;
	case 180: /* 60 GHz band, channels 1..8 */
		if (chan < 1 || chan > 8)
			return -1;
		return 56160 + 2160 * chan;
	case 181: /* 60 GHz band, EDMG CB2, channels 9..15 */
		if (chan < 9 || chan > 15)
			return -1;
		return 56160 + 2160 * (chan - 8);
	case 182: /* 60 GHz band, EDMG CB3, channels 17..22 */
		if (chan < 17 || chan > 22)
			return -1;
		return 56160 + 2160 * (chan - 16);
	case 183: /* 60 GHz band, EDMG CB4, channel 25..29 */
		if (chan < 25 || chan > 29)
			return -1;
		return 56160 + 2160 * (chan - 24);
	}
	return -1;
}

int dpp_parse_uri_chan_list(struct controller *c, struct dpp_enrollee *e, char *uri)
{
	char *pos, *pos2;
	int opclass = -1, channel, freq;

	e->num_chan = 0;

	dbg("|%s:%d| uri:%s\n", __func__, __LINE__, uri);

	pos = strstr(uri, "C:");
	if (!pos) {
		warn("DPP: Invalid URI channel-list");
		return -1;
	}

	pos += 2; /* step past C: */

	while (pos && *pos && *pos != ';') {
		struct dpp_chan *chan;

		if (e->num_chan >= DPP_MAX_CH_OPCLASS)
			break;

		chan = &e->chan[e->num_chan];

		pos2 = pos;
		while (*pos2 >= '0' && *pos2 <= '9')
			pos2++;
		if (*pos2 == '/') {
			opclass = atoi(pos);
			pos = pos2 + 1;
		}
		if (opclass <= 0)
			goto fail;
		channel = atoi(pos);
		if (channel <= 0)
			goto fail;
		while (*pos >= '0' && *pos <= '9')
			pos++;

		freq = ieee80211_chan_to_freq(opclass, channel);
		dbg("DPP: URI channel-list: opclass=%d channel=%d ==> freq=%d",
			   opclass, channel, freq);

		chan->opclass = opclass;
		chan->channel = channel;
		chan->freq = freq;
		e->num_chan++;

		if (*pos == ';' || *pos == '\0')
			break;
		if (*pos != ',')
			goto fail;
		pos++;
	}

fail:
	return 0;
}

uint16_t dpp_get_controller_tcp_port(struct controller *c, enum wifi_band band)
{
	struct dpp_controller_cfg *dpp_cntlr;

	dpp_cntlr = get_dpp_controller_by_band(&c->cfg, band);
	if (dpp_cntlr)
		return dpp_cntlr->port;
	else {
		if (band == BAND_2)
			return 8902;
		else if (band == BAND_5)
			return 8905;
	}

	return 8910;
}

/* will generate an enrollee if it does not exist */
int dpp_append_uri_file(struct controller *c, char *uri)
{
	struct json_object *uri_json, *arr, *val, *new;
	int rv, len, i;

	uri_json = json_object_from_file(DPP_URI_FILE);
	if (!uri_json) {
		dbg("|%s:%d| failed to read json:%s error:%s. "\
		     "Try to generate new\n", __func__, __LINE__, DPP_URI_FILE,
		     json_util_get_last_err());

		uri_json = json_object_new_object();
		if (!uri_json)
			goto out;

		arr = json_object_new_array();
		if (!arr)
			goto out_uri;

		json_object_object_add(uri_json, "uris", arr);
	}

	rv = json_object_object_get_ex(uri_json, "uris", &arr);
	if (!rv || !arr)
		goto out_uri;

	if (!json_object_is_type(arr, json_type_array)) {
		warn("|%s:%d| file:%s is not expected format\n", __func__,
		     __LINE__, DPP_URI_FILE);
		goto out_uri;
	}

	len = json_object_array_length(uri_json);
	for (i = 0; i < len; i++) {
		struct json_object *t;
		const char *p;

		t = json_object_array_get_idx(arr, i);
		if (!t) {
			warn("|%s:%d| couldn't get entry:%d from uri array",
			     __func__, __LINE__, i);
			continue;
		}

		p = json_get_string(t, "uri");
		if (strncmp(p, uri, 128))
			continue;

		/* hash already exists in file */
		/* TODO: possibly type has changed? */
		goto out;
	}

	new = json_object_new_object();
	if (!new)
		goto out_uri;

	val = json_object_new_string(uri);
	if (!val) {
		json_object_put(new);
		goto out_uri;
	}

	json_object_object_add(new, "uri", val);

	val = json_object_new_string("qr");
	if (!val) {
		json_object_put(new);
		goto out_uri;
	}

	json_object_object_add(new, "type", val);
	json_object_array_add(arr, new);
	json_object_to_file(DPP_URI_FILE, uri_json);
out_uri:
	json_object_put(uri_json);
out:
	return 0;
}

/* will generate an enrollee if it does not exist */
struct dpp_enrollee *dpp_process_new_uri(struct controller *c, char *uri)
{
	char *macstr;
	uint8_t enrollee[6] = {0};
	struct dpp_enrollee *e;
	char *hashstr;
	char *hashend;
	int rc;

	if (strncmp(uri, "DPP:", 4)) {
		dbg("|%s:%d| URI is not valid DPP URI!\n", __func__, __LINE__);
		return NULL;
	}

	macstr = strstr(uri, ";M:"); // find mac address in URI
	if (!macstr) {
		macstr = strstr(uri, "DPP:M:"); // find mac address in URI
		if (!macstr) {
			dbg("|%s:%d| URI has invalid format! (Right now) MAC is mandated\n", __func__, __LINE__);
			return NULL;
		}
		macstr += 3; /* step past DPP identifier */
	}

	macstr += 3; /* step past mac identifier */


	if (*(macstr + 12) != ';') {
		dbg("|%s:%d| URI MAC has invalid format!\n", __func__, __LINE__);
		return NULL;
	}

	strtob(macstr, 6, enrollee);

	e = dpp_enrollee_get(c, enrollee);
	if (!e) {
		dbg("|%s:%d| no dpp enrollee for mac "MACFMT", allocate\n",
			__func__, __LINE__, MAC2STR(enrollee));
		e = dpp_enrollee_alloc(c, enrollee);
		if (!e) {
			dbg("|%s:%d| Failed to allocate enrollee for mac "MACFMT"\n",
				__func__, __LINE__, MAC2STR(enrollee));
			return NULL;
		}
	}

	strncpy(e->uri, uri, sizeof(e->uri) - 1);

	hashstr = strstr(uri, ";K:");
	hashstr += 3 + 36; /* TODO: probably step 36 is not correct, but gives correct bytes in hashb output */
	hashend = strstr(hashstr, ";");

	if (hashstr && hashend) {
		uint8_t len = 0;
		uint8_t hashb[128] = {0};
		size_t hlen = 128;

		len = abs(hashend - hashstr);
		rc = base64_decode((unsigned char *) hashstr, (size_t)len, (unsigned char *) hashb, &hlen);
		memcpy(e->hash, hashb, hlen);
		dump(e->hash, hlen, "enrollee hash");
	}

	dbg("|%s:%d| MAC "MACFMT" has URI %s\n", __func__, __LINE__, MAC2STR(e->mac), e->uri);

	send_dpp_cce_indication_all(c, true);

	rc = dpp_parse_uri_chan_list(c, e, e->uri);
	if (rc < 0) {
		dbg("|%s:%d| URI chanlist has invalid format! (Right now) Opclass/Channel list is mandated\n", __func__, __LINE__);
	//	dpp_enrollee_free(e);
	//	return NULL;
	}
	if (e->num_chan) {
		e->band = wifi_channel_to_band(e->chan[0].channel);
		/* need to off-channel listen not yet supported by current
		 * hostapd version used
		 */
	}


//	{
//		/* THIS IS A HACK */
//		uint8_t bssid[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
//		uint8_t radio[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
//		struct cmdu_buff *cmdu;
//		int len = 0;
//
//		len = strlen(e->uri);
//		cmdu = cntlr_gen_dpp_bootstrapping_uri_notification(c, radio, bssid,
//				enrollee, len, e->uri);
//		if (cmdu) {
//			struct node *n;
//
//			list_for_each_entry(n, &c->nodelist, list) {
//				memcpy(cmdu->origin, n->alid, 6);
//				send_cmdu(c, cmdu);
//			}
//			cmdu_free(cmdu);
//		}
//	}

	return e;
}

int dpp_cntlr_read_uris(struct controller *c)
{
	struct blob_buf uris = { 0 };
	struct blob_attr *uri;
	int rem;
	int rc = -1;
	static const struct blobmsg_policy attr[] = {
		[0] = { .name = "uris", .type = BLOBMSG_TYPE_ARRAY },
	};
	struct blob_attr *tb[ARRAY_SIZE(attr)];

	blob_buf_init(&uris, 0);

	if (!blobmsg_add_json_from_file(&uris, DPP_URI_FILE)) {
		warn("Failed to parse %s\n", DPP_URI_FILE);
		goto out;
	}

	blobmsg_parse(attr, ARRAY_SIZE(attr), tb, blob_data(uris.head), blob_len(uris.head));
	if (!tb[0])
		goto out;

	blobmsg_for_each_attr(uri, tb[0], rem) {
		static const struct blobmsg_policy uri_attr[4] = {
			[0] = { .name = "uri", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "type", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *tb1[ARRAY_SIZE(uri_attr)];
		struct dpp_enrollee *e;
		char *uri_str;

		blobmsg_parse(uri_attr, ARRAY_SIZE(uri_attr), tb1, blobmsg_data(uri), blob_len(uri));
		if (!tb1[0] || !tb1[1])
			continue;

		uri_str = blobmsg_data(tb1[0]);
		if (!uri_str)
			continue;

		e = dpp_process_new_uri(c, uri_str);
		if (!e)
			continue;

		if (strcmp("qr", blobmsg_data(tb1[1])))
			e->type = DPP_BOOTSTRAP_QR_CODE_WIFI;
		else
			e->type = DPP_BOOTSTRAP_QR_CODE_WIFI;
	}

out:
	blob_buf_free(&uris);
	return rc;
}

#ifdef USE_LIBDPP
int dpp_frame_handler(void *dpp, uint8_t *smac, enum dpp_event ev,
		      uint8_t *frame, size_t framelen)
{
	struct controller *c;

	err("%s: ----->\n", __func__);

	if (!smac) {
		warn("%s: Error: smac = NULL\n", __func__);
		return -1;
	}

	err("%s: DPP_EVENT: event = %d smac = " MACFMT", frame = %p, len = %zu\n",
		__func__, ev, MAC2STR(smac), frame, framelen);

	switch (ev) {
	case DPP_EVENT_TX_FRAME: {
		struct cmdu_buff *resp;
		int frametype = 0;

		frametype = dpp_get_frame_type(frame + 1, framelen - 1);
		if (frametype == 255) {
			warn("Invalid frametype\n");
			return -1;
		}

		c = dpp_get_ctx_private_data(dpp);
		if (!c)
			return -1;

		dbg("DPP frametype %s\n", dpp_frame_type2str(frametype));
		resp = cntlr_gen_direct_encap_dpp(c, smac, frame, framelen);
		if (!resp) {
			warn("|%s:%d| DPP: ----> Error generating proxy encap\n",
			__func__, __LINE__);
			return -1;
		}

		switch (frametype) {
		case DPP_PA_AUTHENTICATION_REQ: {
#if 0 /* TODO: will be needed for proxied encap */
			struct node *n;

			list_for_each_entry(n, &c->nodelist, list) {
				memcpy(resp->origin, smac, 6);
				send_cmdu(c, resp);
			}
			cmdu_free(resp);

			break;
#else
			/* FALLTHROUGH */
#endif
		}
		case DPP_PA_AUTHENTICATION_CONF:
		case DPP_PUB_AF_GAS_INITIAL_RESP:
		case DPP_PA_PEER_DISCOVERY_RESP:
			memcpy(resp->origin, smac, 6);
			send_cmdu(c, resp);
			cmdu_free(resp);
			break;
		default:
			warn("|%s:%d| DPP: ----> Unknown frame\n", __func__, __LINE__);
			break;
		}
		break;
	}
	case DPP_EVENT_BAD_AUTH:
		err("DPP: BAD AUTH!\n");
		/* FALLTHROUGH */
	case DPP_EVENT_TIMEOUT: {
		struct dpp_peer *peer = dpp_lookup_peer(dpp, smac);

		err("DPP: EVENT TIMEOUT!\n");

		if (peer) {
			list_del(&peer->list);
			dpp_free_peer(dpp, peer);
		}
		break;
	}
	default:
		break;
	}
	return 0;
}
#endif

static void dpp_fd_timeout(atimer_t *t)
{
	struct dpp_chirp *c = container_of(t, struct dpp_chirp, timeout);

	socket_close(&c->chirp_fd.fd);
	uloop_fd_delete(&c->chirp_fd);
}


void dpp_event_uloop_cb(struct uloop_fd *fd, unsigned int events)
{
	struct dpp_chirp *p = container_of(fd, struct dpp_chirp, chirp_fd);
	struct dpp_enrollee *e;
	struct controller *c = p->c; /* TODO: deprecate*/
	struct cmdu_buff *cmdu;
	uint8_t frametype;
	uint8_t *test;
	int len;

	dbg("|%s:%d| Receive event from fd:%d\n", __func__, __LINE__,
	     p->chirp_fd.fd);

	e = dpp_enrollee_get_hash(c, p->hash);
	if (!e) {
		warn("|%s:%d| Get not find enrollee of chirp with an fd!\n",
			__func__, __LINE__);
		/* This is not expected to happen but should probably clear fd here? Close it? */
		return;
	}

	len = dpp_recv_tcp_data(p->chirp_fd.fd, &test);
	if (len == -1) {
		warn("Socket read error, close chirp_fd:%d\n", p->chirp_fd.fd);
		uloop_fd_delete(&p->chirp_fd);
		socket_close(&p->chirp_fd.fd);
		return;
	}

	if (len == 0) {
		warn("No data was read!\n");
		goto out;
	}

	if (FRAME_IS_DPP_PUB_AF(test))
		frametype = dpp_get_pub_af_type(test, len);
	else if (FRAME_IS_DPP_GAS_FRAME(test))
		frametype = dpp_get_gas_frame_type(test, len);
	else
		goto out;

	dump(test, len, "DPP Frame from hostapd");

	cmdu = cntlr_gen_proxied_encap_dpp(c, e->mac, frametype, len, test, 32, p->hash);
	if (cmdu) {
		switch (frametype) {
		case DPP_PA_AUTHENTICATION_REQ: {
			struct node *n;

			list_for_each_entry(n, &c->nodelist, list) {
				memcpy(cmdu->origin, n->alid, 6);
				send_cmdu(c, cmdu);
			}
			cmdu_free(cmdu);
			break;
		}
		case DPP_PA_AUTHENTICATION_CONF: {
			struct dpp_frame *f;

			f = dpp_enrollee_get_frame(c, e, DPP_PA_AUTHENTICATION_RESP);
			if (!f) {
				warn("|%s:%d| Auth resp frame missing! Dropping tcp response\n", __func__, __LINE__);
				break;
			}

			memcpy(cmdu->origin, f->proxy_almac, 6);
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
			break;
		}
		case DPP_PUB_AF_GAS_INITIAL_RESP: {
			struct dpp_frame *f;

			f = dpp_enrollee_get_frame(c, e, DPP_PUB_AF_GAS_INITIAL_REQ);
			if (!f) {
				warn("|%s:%d| Auth resp frame missing! Dropping tcp response\n", __func__, __LINE__);
				break;
			}

			memcpy(cmdu->origin, f->proxy_almac, 6);
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
			break;
		}
		default:
			break;
		}
	}
out:
	free(test);
}
#endif
